/****
*Author Javier Pena
*Date April 30, 2016
*
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <errno.h>
#include "list.h"
#include "chop_line.h"
#include "parser.h"


int fileExists(char *file)
{
  if(access(file,F_OK) != -1)
	return 1;
  else
	return 0;


}
/*
  handle a single right redirect.
 */
int redirectR(CMD_t *cmd, char *cmdstr, char **args)
{
  int file;
  /* printf("creating new file\n"); */
  file = open(cmd->fname,O_CREAT | O_EXCL | O_RDWR,0644);
  if(file > 0)
	{
	  if(dup2(file, 1) > 0)
		{
		  close(file);
		  execvp(cmdstr,args);
		  fprintf(stderr,"execvp file redirection failed\n");
		  exit(-1);
		}
	  else
		{
		  fprintf(stderr,"dup2> failed");
		  exit(-1);
		}
	}
  else
	{
	  fprintf(stderr,"open failed\n");
	  exit(-1);
	}
}

/*
  handle a double right redirect
 */
int redirectRR(CMD_t *cmd, char *cmdstr, char **args)
{
  int file;
  /* printf(">> command\n"); */

 
  if(fileExists(cmd->fname))
	{
	  /* printf("opening file %s\n",cmd->fname); */
	  file = open(cmd->fname, O_RDWR|O_APPEND,0644);
	}
  else
	{
	  file = open(cmd->fname, O_RDWR|O_CREAT,0644);
	}
	  if(file > 0)
		{
		  if(dup2(file, 1) > 0)
			{
			  close(file);
			  execvp(cmdstr,args);
			  fprintf(stderr,"execvp file redirection failed\n");
			  exit(-1);
			}
		  else
			{
			  fprintf(stderr,"dup2>> failed\n");
			  exit(-1);
			}
		}
	  else
		{
		  fprintf(stderr,"open failed\n");
		  exit(-1);
		}


}
/*handle a left redirect
 */
int redirectL(CMD_t *cmd, char *cmdstr, char **args)
{
  int file;
  /* printf("< command\n"); */

  if(fileExists(cmd->fname))
	{
	  /* printf("opening file %s\n",cmd->fname); */
	  file = open(cmd->fname, O_RDONLY);
	  if(file > 0)
		{
		  if(dup2(file, 0) >= 0)
			{
			  close(file);
			  execvp(cmdstr,args);
			  fprintf(stderr,"execvp file redirection failed\n");
			  exit(-1);
			}
		  else
			{
			  int err = errno;
			  fprintf(stderr,"dup2< failed with\n%s\n",strerror(err));
			  exit(-1);
			}
		}
	  else
		{
		  fprintf(stderr,"open failed\n");
		  exit(-1);
		}
	}
  else
	{
	  fprintf(stderr,"file does not exist!");
	}

  
}

/*
  function used to determine which redirect is used and pass it off
  to one of the helper functions
 */
int redirect(CMD_t *cmd, char *cmdstr, char **args)
{

  char *op = cmd->operators;
  /* printf("operator = %s\n",op); */
  if(!strcmp(op,">"))
	{
	  redirectR(cmd,cmdstr,args);
	}
  else if(!strcmp(op,">>"))
	{
	  redirectRR(cmd,cmdstr,args);
	}
  else if(!strcmp(op,"<"))
	{
	  redirectL(cmd,cmdstr,args);
	}
  else
	{
	  fprintf(stderr,"not a file redirect\n");
	  exit(-1);

	}
}

list_t *reverseList(list_t *cmdList)
{
  list_t *reversed = list_create();
  struct list_node_t *curNode = cmdList->head;
  do
	{
	  list_insert_val(reversed,curNode->val);
	}
  while((curNode = curNode->next)!= NULL);

  return reversed;
}

/*
  handle piped commands
*/
int pipeCMD(list_t *cmdList, int listSize)
{
  int numPipes = listSize-1;
  int p = 0;
  int fd[2];
  int status;
  int pipeResult;
  pid_t chd[listSize];
  /*reversing the list is necessary because when inserting into the list
	we end up reversing the order of commands from how we were originally given
   */
  list_t *revCmdList = reverseList(cmdList);
  struct list_node_t *curNode = revCmdList->head;
  char *executable;
  char **args;
  /* printf("piping %d commands\n",listSize); */
  pipeResult = pipe(fd);
  do
	{
	  executable = curNode->val->executable;
	  args = curNode->val->args;	  
	  if((chd[p] = fork()) == 0)
		{
		  if(p==0)//handle the first command
			{
			  dup2(fd[1],1);
			  close(fd[0]);
			  close(fd[1]);
			  execvp(executable, args);
			  exit(-1);
			}
		  else if(p==numPipes)//handle the last command
			{
			  dup2(fd[0],0);
			  close(fd[0]);
			  close(fd[1]);
			  if(curNode->val->hasRedirect)
				{
				  redirect(curNode->val, executable, args);
				  exit(-1);
				}
			  execvp(executable, args);
			  exit(-1);
			}
		  else//handle the inbetween commands
			{
			  dup2(fd[0],0);
			  dup2(fd[1],1);
			  close(fd[0]);
			  close(fd[1]);
			  execvp(executable, args);
			  exit(-1);
			}
		}
	  p++;
	}
  while((curNode = curNode->next) != NULL);
  close(fd[0]);
  close(fd[1]);
  int i;
  for(i = 0; i<listSize;i++)
	{
	  waitpid(chd[i],&status,0);
	  /* if(WIFEXITED(status)) */
	  /* 	printf("PARSERchild %d terminated with status %d\n",chd[i],WEXITSTATUS(status)); */
	  /* else */
	  /* 	printf("PARSERchild %d terminated abnormally\n",chd[i]); */
	  if(!WIFEXITED(status))
		fprintf(stderr,"child %d terminated abnormally\n",chd[i]);
	}
}

int executeCMD(list_t *cmdList, int listSize)
{
  /* printf("listSize in executeCMD = %d\n",listSize); */
  CMD_t *cmd = cmdList->head->val;
  char* cmdstr = cmd->executable;
  char** args = cmd->args;;
  int status;
  /* printf("executing command %s\n",cmd); */
  /* printf("num used %d\n",curNode->val->used); */
  int i = 0;
  pid_t child;
  
  if((child = fork()) == 0)
	{
	  /* printf("executing command\n"); */
	  //	  printf("in child parser - process group = %d\n",getpgrp());
	  if(listSize > 1)// we only insert commands into a list when they are piped
		{
		  pipeCMD(cmdList,listSize);
		}
	  else if(cmd->hasRedirect)
		{
		  if(cmd->hasFile)
			{
			  /* printf("redirecting file\n"); */
			  redirect(cmd, cmdstr, args);
			}
		  else
			{
		  fprintf(stderr,"error missing file for redirect\n");
			  exit(-1);
			}
		}
	  else
		{
		  execvp(cmdstr,args);
		  exit(-1);
		}
	}
  else
	{
	  // printf("in parent parser - process group = %d\n",getpgrp());
	  /* printf("waiting for process %d\n",child); */
	  while(errno != 3)
	  {
		kill(child, 0);
	  }
	  return 0;
	}
}
