/****
*Author Javier Pena
*Date April 30, 2016
*
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/wait.h>
#include <fcntl.h>
#include <errno.h>
#include "list.h"
#include "chop_line.h"
#include "parser.h"


/*
  Todo:
  need to wait for bg commands to terminate before handling EOF
  need to fix simsh prompt printing too fast in some instances
*/

#define TRUE 1;
#define FALSE 0;
int maxChar = 4096;

/**
   not yet used or fully implemented
 */
/* void freeCMD(CMD_t *cmd) */
/* { */
/*   int xCounter = 0; */
/*   int argCounter = 0; */
/*   int opCounter = 0; */
/* } */

/**
   All of these get commands take the current token and place them
   into a CMD_t struct(located in list.h) in their proper place.
 */
int getCommand(CMD_t *cmd, int curToken, chopped_line_t *cl)
{
  cmd->executable = strdup(cl->tokens[curToken]);
  getArg(cmd,curToken,cl);
  return 0;
}

int getFile(CMD_t *cmd, int curToken, chopped_line_t *cl)
{
  cmd->fname = strdup(cl->tokens[curToken]);
  /* printf("%s\n",cmd->fname); */
  return 0;
}

int getArg(CMD_t *cmd, int curToken, chopped_line_t *cl)
{
  if(cmd->usedArgs == cmd->sizeArgs)
	{
	  cmd->sizeArgs += 8;
	  cmd->args = realloc(cmd->args,sizeof(cmd->sizeArgs) * sizeof(CMD_t *));
	}
  cmd->args[cmd->usedArgs++] = strdup(cl->tokens[curToken]);
  //  printf("%s\n",cl->tokens[curToken]);
  return 0;
}

int getOperator(CMD_t *cmd, chopped_line_t *cl, int curToken,char *op)
{
  /* printf("getOperator --- operators = %s\n",cmd->operators); */
  /* printf("getOperator --- op =%s\n",op); */
  cmd->hasOp = TRUE;

  //if the operator is a pipe DO NOT copy it.
  if(strcmp(op, "|"))  cmd->operators = strdup(op);

  return 0;
}

int isOperator(char* token)
{

  if(!strcmp(token, "|"))
	{
	  /* printf("pipe found!    \n"); */
	  /* op = strdup(token); */
	  return TRUE;
	}
  else if(!strcmp(token,"&"))
	{
	  /* printf("ampersand found!   \n"); */
	  /* op = strdup(token); */
	  return TRUE;
	}
  else if(!strcmp(token,"<"))
	{
	  /* printf("left redirect found!   \n"); */
	  /* op = strdup(token); */
	  return TRUE;
	}
  else if(!strcmp(token,">>"))
	{
	  /* printf("right double redirect found!   \n"); */
	  /* op = strdup(token); */
	  return TRUE;
	}
  else if(!strcmp(token,">"))
	{
	  /* printf("right redirect found!   \n"); */
	  /* op = strdup(token); */
	  /* printf("isOperator = %s\n",op); */
	  return TRUE;
	}
  else
	{
	  return FALSE;
	}
}


CMD_t *initCommand()
{
  CMD_t *cmd = malloc(sizeof(CMD_t));
  cmd->sizeArgs=0;
  cmd->usedArgs=0;
  cmd->hasArgs=FALSE;
  cmd->hasOp=FALSE;
  cmd->hasFile=FALSE;
  cmd->hasRedirect=FALSE;
  cmd->bg=FALSE;
  return cmd;
}
/**
   parse the chopped line containing all tokens from stdin
   and place them into a CMD_t struct which will them be placed into a
   list. The list is mainly used for piping commands together.
*/
int parse(list_t *cmdList, chopped_line_t *cl)
{

  int curToken = 0;
  char *token;
  int commandSearch = TRUE;
  int argSearch = FALSE;
  int fileSearch = FALSE;
  int ambiguous = FALSE;
  CMD_t *cmd;
  char *op;
  int listSize = 0;

  /* printf("num tokens = %d\n",cl->num_tokens); */
  
  while(curToken < cl->num_tokens)
	{
	  token = cl->tokens[curToken];
	  if(commandSearch)
		{ 
		  /* printf(">>>grabbing command === "); */
		  cmd = initCommand();
		  list_insert_val(cmdList,cmd);
		  getCommand(cmd, curToken, cl);
		  if(isOperator(cmd->executable))
			{
			  fprintf(stderr,"Missing Command!");
			  return -1;
			}

		  commandSearch = FALSE;
		  argSearch = TRUE;
		  listSize++;
		}
	  else if(argSearch)
		{
		  if(isOperator(token))
			{
			  /* printf("++++grabbing op = %s\n",strdup(token)); */
			  op = strdup(token);
			  getOperator(cmd, cl, curToken, op);
			  /* printf("op retrieved = %s\n",op); */
			  if(!strcmp(token, ">") || !strcmp(token, ">>") || !strcmp(token,"<"))
				{
				  ambiguous = TRUE;//used to help determing ambiguous output
				}
			  
			  if(!strcmp(token, "|"))
				{
				  if(ambiguous)
					{
					  fprintf(stderr,"Ambiguous output redirect.\n");
					  return -1;
					}
				  commandSearch = TRUE;
				}
			  else if(!strcmp(token, "&"))
				{
				  if(curToken == (cl->num_tokens - 1))
					{
					  cmd->bg = TRUE;
					}
				  /* if(fileSearch) */
				  /* 	{ */
 				  /* 	  fileSearch = FALSE; */
				  /* 	  argSearch = FALSE; */
				  /* 	  cmd->bg = TRUE; */
				  /* 	} */
				}
			  else
				{
				  /*
					It must be a file redirect since we already checked if the
					token is a valid operator.
				   */
				  cmd->hasRedirect = TRUE;
				  fileSearch = TRUE;
 				  argSearch = FALSE;
				}
			}
		  else
			{
			  if(!fileSearch)
				{
				  getArg(cmd,curToken,cl);
				  cmd->hasArgs = TRUE;
				}
			}
		}
	  else if(fileSearch)
		{
		  /* printf("***grabbing file = "); */
		  getFile(cmd,curToken,cl);
		  cmd->hasFile = TRUE;
		  argSearch = TRUE;
		}
	  curToken++;
	}
  return listSize;
}

chopped_line_t *readLine()
{
  char *str = malloc(4097);
  unsigned charCount = 0;
  char c;

  do
	{
	  c = getchar();
	  if(c == EOF)
		exit(0);
	  str[charCount++] = c;
	}
  while(charCount < 4096 && c !='\n');
  str[charCount] = '\0';
  
  return get_chopped_line(str);
}




